package no.ntnu.stockmeks;

import org.jgap.FitnessFunction;
import org.jgap.IChromosome;
import no.ntnu.stockmeks.agent.*;

/*
 * Class responsible for evaluating a chromosome with a fitness function
 */
public class GeneticFitness extends FitnessFunction {
	
	private static final long serialVersionUID = 1L;
	StockTicker[] tickers;
	Agent[] agents;
	AgentEvaluator ae;
	AgentAdministrator[][] aa;
	StockTickerExtremes[] yield;
	
	double buySellEarnings = 0.0;
	long buySellTransactions = 0;

	
	int mode = 0;

	
	
	public GeneticFitness(StockTicker[] tickers, Agent[] agents,int mode){
		this.mode = mode;
		if(mode>0) initializeWithExtremes(tickers,agents);	 
		else initialize(tickers,agents);
	}

	
	
	public void initialize(StockTicker[] tickers, Agent[] agents){
		aa = new AgentAdministrator[tickers.length][agents.length];
		
		for(int i=0;i<tickers.length;i++){
			for(int j=0;j<agents.length;j++){
				aa[i][j] = new AgentAdministrator(tickers[i],agents[j]);
			}
		}
		this.agents = agents;
		this.tickers = tickers;
	}
	
	
	public void initializeWithExtremes(StockTicker[] tickers, Agent[] agents){
		aa = new AgentAdministrator[tickers.length][agents.length];
		yield = new StockTickerExtremes[tickers.length];
		
		for(int i=0;i<tickers.length;i++){
			yield[i] = new StockTickerExtremes(tickers[i],5.0);
			for(int j=0;j<agents.length;j++){
				aa[i][j] = new AgentAdministrator(tickers[i],agents[j]);
			}
		}
		this.agents = agents;
		this.tickers = tickers;
	}
	
	
	/*
	 * Fitness function calculations called by genetics main method 
	 */
	protected double evaluate(IChromosome kromosom) {
		
		double totalFitness = 0.0;
		buySellEarnings = 0.0;
		buySellTransactions = 0;
		
		int[] allele = new int[kromosom.size()];
		int totalAllele = 0;
		for(int i=0;i<kromosom.size();i++){
			allele[i] = (Integer)kromosom.getGene(i).getAllele();
			if(allele[i]<0) totalAllele += (allele[i] *(-1));
			else if(allele[i]>0) totalAllele += allele[i];
		}
		
		
		for(int i=0;i<tickers.length;i++){
			double[] agentSignals = new double[tickers[i].getSize()];
			for(int day=0;day<tickers[i].getSize();day++){
				int alleleSignal = 0;
				int alleleAtDay = 0;
				for(int k=0;k<agents.length;k++){
					int agentSignalAtDay = aa[i][k].getAgentSignal(day);
					alleleSignal += (allele[k]*agentSignalAtDay);
					if(agentSignalAtDay!=0){
						if(allele[k]>0) alleleAtDay += allele[k];
						else alleleAtDay += (-1)*allele[k];
					}
				}
				agentSignals[day] = (double)alleleSignal/(double)alleleAtDay;
			}

			
			
			//have calculated all signal for all days
			if(mode==1) totalFitness += getOptimalYield(i,agentSignals);
			else if(mode==2) totalFitness += (getCorrectDays(i,agentSignals)/(double)tickers.length);
			else if(mode==3) totalFitness += getBuySellYield(tickers[i],agentSignals);
			else totalFitness += getYield(tickers[i],agentSignals);
		}
		
		if(totalFitness<0) return 0.0;
		else return totalFitness;
	}

	
	
	public double getBuySellYield(StockTicker ticker,double[] signals){
		double boughtAt = 0;
		double soldAt = 0;
		
		boolean doShortSell = true;
		int startDay = 30;
		
		double totalEarnings = 0;
		int buys = 0;
		int sells = 0;
		
		for(int i=startDay;i<ticker.getSize();i++){
			
			if(boughtAt==0 && signals[i]>0.5){
				//want to buy
				buys++;
				if(soldAt==0){
					boughtAt = ticker.getClose(i);
				}
				else{
					double transactionEarnings = 1.0-((ticker.getClose(i)/soldAt));

					buySellEarnings+=transactionEarnings;
					buySellTransactions++;

					totalEarnings+=transactionEarnings;
					
					soldAt=0;
					boughtAt = ticker.getClose(i);
				}
				
			}
			else if(soldAt==0 && signals[i]<-0.5){
				sells++;
				if(boughtAt==0){
					//short selling
					if(doShortSell){
						soldAt = ticker.getClose(i);
					}
				}
				else{
					double transactionEarnings = ((ticker.getClose(i)-boughtAt)/boughtAt);
					
					buySellEarnings+=transactionEarnings;
					buySellTransactions++;

					totalEarnings+=transactionEarnings;
					boughtAt = 0;
					if(doShortSell) soldAt = ticker.getClose(i);
				}
				
				
			}
			
		}
		
		if(soldAt>0){
			double transactionEarnings = 1.0-((ticker.getClose(ticker.getSize()-1)/soldAt));
			buySellEarnings+=transactionEarnings;
			buySellTransactions++;
			totalEarnings+=transactionEarnings;

		}
		else if(boughtAt>0){
			double transactionEarnings = ((ticker.getClose(ticker.getSize()-1)-boughtAt)/boughtAt);
			buySellEarnings+=transactionEarnings;
			buySellTransactions++;

		}
	
		return totalEarnings;
	}
	
	
	public double getCorrectDays(int tickerNumber,double[] signals){
		int totalCorrect = 0;
		double above = 1.01;
		double below = 0.99;
		
		
		for(int i=0;i<signals.length;i++){
			if(signals[i]>0.5){
				if(yield[tickerNumber].getExtremeYield(i)>above) totalCorrect++;
			}
			else if(signals[i]<-0.5){
				if(yield[tickerNumber].getExtremeYield(i)<below) totalCorrect++;
			}
			else if(yield[tickerNumber].getExtremeYield(i)>below && yield[tickerNumber].getExtremeYield(i)<above) totalCorrect++;
		}
		
		return ((double)totalCorrect/(double)tickers[tickerNumber].getSize());
	}
	
	public double getOptimalYield(int tickerNumber,double[] signals){
		
		double totalFitness = 0.0;
		
		for(int i=0;i<signals.length;i++){
			if(signals[i]>0.5){
				totalFitness += yield[tickerNumber].getExtremeYield(i);
			}
			else if(signals[i]<-0.5){
				if(yield[tickerNumber].getExtremeYield(i)>0){
					totalFitness += (1.0)/(yield[tickerNumber].getExtremeYield(i));
				}
			}
		}
		return totalFitness;
	}
	
	
	public double getYield(StockTicker ticker,double[] signals){
		double boughtAt = 0.0;
		double soldAt = 0.0;
		double earnings = 0.0;
		
		int transactions = 0;

		
		for(int i=0;i<signals.length;i++){
			//simulating buy and sell each day
			if(boughtAt==0 && signals[i]>0.5){
				//want to buy
				boughtAt = ticker.getClose(i);
				transactions++;

				if(soldAt>0){
					//have short sold earlier, buying back
					double transactionEarnings = 1.0-((ticker.getClose(i)/soldAt));
					//earnings+=(soldAt-ticker.getClose(i))/soldAt;

					earnings += transactionEarnings;
					soldAt=0;
				}	
				
			}
			else if(soldAt==0 && signals[i]<-0.5){
				soldAt = ticker.getClose(i);
				transactions++;
				
				if(boughtAt>0){
					//have bought earlier, selling this
					double transactionEarnings = ((ticker.getClose(i)-boughtAt)/boughtAt);
					earnings+= transactionEarnings;
					
					boughtAt = 0;
				}
			}
			
		}
		if(transactions>0) return earnings/(double)transactions;
		else return 0.0;
	}
	
	
}
